/* -*- mode: c++; c-basic-offset: 3; -*- */
//
//    DPUSHRT is the application to push the data from a file directory 
//    into the shared memory partition in frame format. 
//
//    Syntax:
//
//       DpushRT {infile <file> | inlist <list-file>} [partition <pname>] 
//               [lbuf <nbytes>] [nbuf <n>] [log <logfile>] [time <ticks>]
//
//       file      Wildcard list of files to be pushed.
//       list-file File containing a list of input files.
//       pname     Shared memory partition name (defaults to $LIGOSMPART)
//       nbytes    Length of shared memory buffers
//       n         Number of shared memory buffers
//       logfile   Log file.
//       ticks    Watchdog (statistics summary) timeout interval
//
//    Notes:
//      1) Partitions aren't currently created by the program.
//
//    Revision History:
//
//      02-Feb-2021  J. Zweizig
//          Preliminary version.
//
//------- A few definitions
//
//    DEFNDSIP the default NDS server IP address.
//    DEFPART  is the default partition name that is used if the "partition" 
//             command line argument is not specified.
//    CDSDATA  receive data in CDS proprietary format.
//
// #define _POSIX_C_SOURCES
#if defined(sun) && !defined(__EXTENSIONS__)
#define __EXTENSIONS__
#endif

//  Use CDSDATA to disable frame checking.
#define CDSDATA 1
#define DEFPART "LIGO_Offline"

//------- Include till you drop.
#include "PConfig.h"
#include <iostream>
#include <fstream>
#include <sstream> 
#include <cstdlib>
#include <unistd.h>
#include <string>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include "SigFlag.hh"
#include "SysError.hh"
#include "Time.hh"
#include "lsmp_prod.hh"
#include "framedir.hh"
#include "CheckFrame.hh"
#include "time_utils.hh"
#include <memory>

using namespace std;

//------- Statistics class.
class Stats {
public:
   typedef unsigned long ulong_t;
   Stats();
   ~Stats();
   void Log(ostream& out);
   void Reset(void);
   void Start(ulong_t Secs);
   bool Test(void);
   void Update(ulong_t FrameID, const Time& GPS, ostream& out);
private:
   ulong_t nSecs;
   SigFlag watchdog;
   Time    Stamp;
   Time    GPSstart;
   Time    GPSlast;
   ulong_t RecIDstart;
   ulong_t RecIDlast;
   ulong_t FrameCount;
   ulong_t FrameLost;
   ulong_t FrameUnsent;
};

class DpushRT {
public:
   ///  Constructor
   DpushRT(int argc, const char* argv[]);
   ///  Destructor
   ~DpushRT(void);
   ///  initialization - Set up the input and output functions
   bool Init(int argc, const char* argv[]);
   ///  Loop over frames until the end of time.
   void Loop(void);
   ///  Read a file and copy it to the 
   bool Read(void);
   ///  Check the frame validity.
   bool Check(unsigned int length, const char* addr);
   ///  Termination method - called from the destructor
   void Halt(void);

private:
   bool     active;    ///<  run data copying.
   bool     debug;     ///<  Print debugging information
   bool     LogOn;     ///<  Enable error logging
   double   compress;  ///<  replay speed-up ratio
   Interval delay;     ///<  replay delay
   Time     startTime; ///<  replay starting time.
   Interval allowance; ///<  maximum delay allowance for late frames

   //---->  Time control
   std::unique_ptr<LSMP_PROD> OPart;  ///<  Shared memory output device
   SigFlag term;       ///<  Termination signal flag pointer
   unsigned long FrameID;
   FrameDir fDir;
   Stats   pgmStat;    ///<  Watchdog statistics structure.
   ofstream logfi;
   FrameDir::file_iterator pFrame;
   Time     nextTime;  ///<  Time to replay the next frame
};

//======================================  Static functions
string stamp(time_t now);

//
//    main function. 
//      * Set up. This includes initializing the input and output 
//        files and the exception handling.
//      * Loop over records until the ceiling falls in on us
//      * Call Read to process a frame
//      * When the roof caves in set-down the application
//
int main(int argc, const char **argv) {

   DpushRT dpush(argc, argv);

   dpush.Loop();

   return 0;
}

//======================================  Constructor
DpushRT::DpushRT(int argc, const char **argv)
   : debug(false), LogOn(false), compress(1.0), delay(0.0), allowance(1.0)
{
   active = !Init(argc, argv);
}

//======================================  Destructor
DpushRT::~DpushRT(void) {
   Halt();
}

//
//    Init - set up the internal environment for pushing data.
//       * Parse any command line parameters.
//       * Create a shared memory partition for data distribution.
//       * Make a connection to the CDS Network Data Server. Register 
//         notification of all frames.
//       * Set up the termination signal handling.
//
bool DpushRT::Init(int argc, const char* argv[]) {
   int   nbuf = 1, lbuf = 1000000;

   //---------------------------------  Define default parameters
   string ParName;
   const char* partenv = getenv("LIGOSMPART");
   if (partenv) ParName = partenv;
   else         ParName = DEFPART;

   //---------------------------------  Parse the arguments
   for (int i=1 ; i<argc ; i++) {
      string argi = argv[i];
      //------------------------------  Specify output partition.
      if (argi == "-partition") {
	 ParName  = argv[++i];
      }

      //------------------------------  Specify input file(s) (globs)
      else if (argi == "-infile") {
	 fDir.add(argv[++i]);
      }

      //------------------------------  File containing a list of input files.
      else if (argi == "-inlist") {
	 ifstream list(argv[++i]);
	 if (list.is_open()) {
	    std::string line;
	    while (getline(list, line)) {
	       line.erase(0, line.find_first_not_of(" \t"));
	       if (line[0] == '#') continue;
	       line.erase(line.find_last_not_of(" \t")+1);
	       fDir.add(line.c_str());
	    }
	 } else {
	    cerr << "DpushRT: Unable to open list file: " << argv[i]  << endl;
	 }
      }

      //--------------------------------  Partition buffer length and number
      else if (argi == "-lbuf") {
	 lbuf = strtol(argv[++i], 0, 0);
      }
      else if (argi == "-nbuf") {
	 nbuf = strtol(argv[++i], 0, 0);
      }

      //--------------------------------  Replay start time
      else if (argi == "-time") {
	 Time::ulong_t Secs = 0;
	 if (i+1 < argc && *argv[i+1] != '-') Secs = strtol(argv[++i], 0, 0);
	 if (Secs) startTime = Time(Secs);
	 else      startTime = Now();
      }

      //--------------------------------  Time compression factor 
      else if (argi == "-compress") {
	 compress = strtod(argv[++i], 0);
      }

      //--------------------------------  debugging printout flag
      else if (argi == "-debug") {
	 debug = true;
      }

      //-------------------------------  Allow replay to be data delayed.
      else if (argi == "-allow") {
	 allowance = Interval(strtod(argv[++i], 0));
      }
      //-------------------------------  Absolute delay from frame time stamp
      else if (argi == "-delay") {
	 delay = Interval(strtod(argv[++i], 0));
      }

      //-------------------------------  Specify and create a log file.
      else if (argi == "-log") {
	 LogOn = true;
	 string lfile;
	 if (i+1 >= argc || !*argv[i+1] || *argv[i+1] == '-') {
	    lfile  = getenv("HOME");
	    lfile += "/logs/";
	    lfile += LocalString(Now(), "DpushRT-%Y.%02m.%02d-%02H.%02N.log");
	 } else {
	    i++;
	    lfile  = argv[i];
	 }
	 logfi.open(lfile.c_str(), ios::out);
	 if (logfi.good()) {
	    logfi << stamp(0) << argv[0] << " Starting..." 
		  << endl;
	 } else {
	    LogOn = false;
	 }
      }

      //--------------------------------  Unrecognized command line argument.
      else {
	 cerr << argv[0] << ": Invalid argument name " << argv[i] 
	      << endl;
	 cerr << "Command syntax:" << endl << endl;
	 cerr << argv[0] << " {-infile <file> | -inlist <list-file>} "
	                    "[-partition <partition>] \\ " << endl
	      << "        [-nbuf <#buffers>] [-lbuf <buffer-length>] "
	                 "[-log <log-file>] \\ " << endl
	      << "        [-time <start-time>] [-compress <factor>] "
	                 "[-delay <delta-t>] \\ " << endl
	      << "        [-allow <allowance>] [-debug]" << endl;
	 return true;
      }
   }

   //---------------------------------  Open the output partition.
   if (debug) cout << "Opening partition " << ParName << " for output." 
		   << endl;
   OPart.reset(new LSMP_PROD(ParName.c_str(), nbuf, lbuf));
   if (!OPart->valid()) {
      cerr << "Unable to access partition " << ParName << ": " 
	   << OPart->Error() << "." << endl;
      if (LogOn) logfi << stamp(0) << "Unable to access partition " 
		       << ParName << ": " << OPart->Error() << "." 
		       << endl;
      return true;
   }

   //---------------------------------  Set up the statistics task
   if (LogOn) pgmStat.Start(startTime.getS());

   //---------------------------------  Set up the termination signals
   term.add(SIGINT);
   term.add(SIGTERM);
   term.add(SIGHUP);

   //---------------------------------  Get the start time;
   FrameID = 0;

   //---------------------------------  Skip past expired frames
   pFrame = fDir.begin();

   if (!startTime) {
      Time now(Now());
      for (; pFrame != fDir.end(); pFrame++) {
	 if (pFrame->getEndTime() >= now + delay + allowance) break;
      }
      if (pFrame != fDir.end()) {
	 cout << "DpushRT: Skipped frames to " << pFrame->getStartTime().getS()
	      << endl;
      }
   }

   //---------------------------------  Done, quit.
   return false;
}

//
//    Halt - set down.
//       * Release the shared memory partition.
//       * Close the connection to the CDS Network Data Server.
//       * Cancel the termination signal handling.
//
void DpushRT::Halt(void) {

   //---------------------------------  Disable the signal
   term.zero();

   //---------------------------------  Final statistics
   pgmStat.Log(logfi);

   //---------------------------------  Close the error log
   if (LogOn) {
      LogOn = false;
      logfi << stamp(0) << "Dpush terminating. Error log closed" 
	    << endl;
      logfi.close();
   }
}

//======================================  Read - read the data.
//
//      *  Wait for a frame to arrive.
//      *  Read in the data
//      *  Pack data into a frame
//      *  Write frame to shared memory.
//
bool
DpushRT::Read(void) {
   //---------------------------------- Wait for a data message
   if (pFrame == fDir.end()) {
      cout << "Stopping after last file" << endl;
      active = false;
      return false;
   }

   //----------------------------------  Open the file
   Time GPS = pFrame->getStartTime();
   Interval FrameDt =  pFrame->getDt();
   string fName = pFrame->getFile();
   int fd = open(fName.c_str(), O_RDONLY);
   if (fd < 0) {
      cerr << "Error opening file: " << fName << endl;
      return false;
   } else if (debug) {
      cout << "Opened frame file: " << fName << endl;
   }

   //----------------------------------  Get the file length - check buffer
   struct stat statBuf;
   fstat(fd, &statBuf);
   long lFile = statBuf.st_size;

   if (lFile > OPart->getBufferLength()) {
      active = false;
      cerr << "Frame longer than shared memory buffer." << endl;
      if (LogOn) logfi << stamp(0) << "Frame longer than shared memory buffer"
		       << endl;
      return false;
   }

   //----------------------------------  Get a buffer
   char* buffer = reinterpret_cast<char*>(0);
   while (!buffer) {
      try {
	 buffer = OPart->get_buffer();
      } catch (exception& e) {
	 active = false;
	 cerr << "Error allocating buffer:" << e.what() << endl;
	 if (LogOn) logfi << stamp(0) << "Error allocating buffer:" 
			  << e.what() << endl;
         close(fd);
	 return false;
      }	    
   }

   //---------------------------------  Calculate nominal replay time
   if (startTime) {
      if (!nextTime) nextTime = startTime;
      nextTime += FrameDt * compress;
   } else {
      nextTime = GPS + FrameDt;
   }

   //---------------------------------  Wait for replay time
   Time tNext = nextTime + delay;
   Time now = Now();
   if (tNext > now) {
      if (debug) cout << "DpushRT: Waiting " << tNext - Now() << "sec." << endl;
      sleep_until(tNext);
   }
   
   //----------------------------------  drop the frame if too late
   else if (tNext + allowance > now) {
      cerr << "Frame (" << fName << ") read after allowance expired. "
	   << "wait time = " << tNext - now << endl;
      OPart->return_buffer();
      close(fd);
      FrameID++;
      return false;	
   }

   //-----------------------------------  Warn allowance used if debuging.
   else if (debug) {
      cout << "Frame (" << fName << ") delayed within allowance (dt="
           << tNext - now << ")." << endl;
   }

   //---------------------------------  Copy the data
   long length = read(fd, buffer, lFile);
   if (debug) cout << "Read frame: file-length/rc " << lFile
		   << " / " << length << endl;
   close(fd);
   FrameID++;

   if (length == 0) {
      OPart->return_buffer();
      cerr << "Read error." << endl;
      if (LogOn) logfi << stamp(0) << "Read error." << endl;
      active = false;
   } else if (!Check(length, buffer)) {
      OPart->return_buffer();
      if (LogOn) {
	 logfi << stamp(0) << "Data check failed on frame " << FrameID 
	       << " GPS: " << GPS.getS() << ")" << endl;
      } else {
	 cerr << "Data check failed on frame " << FrameID 
	       << " GPS: " << GPS.getS() << ")" << endl;
      }
   } else {
      OPart->SetID(GPS.getS());
      OPart->release(length);
   }

   //----------------------------------  Check for discontinuities
   if (LogOn) pgmStat.Update(FrameID, GPS, logfi);
   return true;
}

//======================================  Loop over the read/write operations
void DpushRT::Loop(void) {
   while (active && !term) {
      if (LogOn && pgmStat.Test()) pgmStat.Log(logfi);
      bool ok = Read();
      pFrame++;
   }
}

//======================================  Frame Check Code
//
//  Check  Checks the incoming frame data structure headers for 
//              consistency.
//
//  The structure headers must meet the following requirements
//  1) The length must be greater than 8 (the length of the header) and 
//     less than the length of the Frame.
//  2) The structure type must be 1 or 2 or have been defined in a previous
//     SH structure. The Structure ID must be in the range 0 < ID < MAXID 
//     where MAXID is arbitrarily set to 256 (for now).
//  3) the structure instance numbers must increase by one with each new 
//     instance and start at either 0 or 1.
//
//  In addition the SH structures are unpacked and must be internally 
//  consistent (see procSH).
//
//  Check returns true if the frame passes the requirements.
//
class DPCheck : public CheckFrame {
public:
   DPCheck(istream& s);
   void ErrorOut(const string& msg);
   void set_log(std::ostream* log);
private:
   std::ostream* logfi;
};

DPCheck::DPCheck(istream& s) 
   :  CheckFrame(s), logfi(0) {
}

void 
DPCheck::ErrorOut(const string& s) {
   if (logfi) (*logfi) << stamp(0) << s << endl;
   cerr << s << endl;
}

void
DPCheck::set_log(std::ostream* log) {
   logfi = log;
}

bool
DpushRT::Check(unsigned int length, const char* addr) {
#ifndef CDSDATA
   istringstream s(string(addr, length));
   DPCheck c(s);
   if (LogOn) c.set_log(logfi);
   try {
      c.Check();
   } catch (exception& e) {
      if (LogOn) logfi << stamp(0) << e.what() << endl;
      cerr << e.what() << endl;
      return false;
   }
#endif  // !def(CDSDATA)
   return true;
}

string
stamp(time_t now) {
   if (now == 0) now = time(0);
   string datel(::ctime(&now)+4, 15);
   datel += " ";
   return datel;
}

//--------------------------------------  Statistics class methods
Stats::Stats(void) 
   : RecIDlast(0)
{}

Stats::~Stats() {
}

bool 
Stats::Test() {
   return watchdog;
}

void 
Stats::Log(ostream& out) {
   Time   now = Now();
   double Rate;
   if (now == Stamp) Rate = 1.0;
   else              Rate = (GPSlast - GPSstart) / (now-Stamp);
   if (FrameLost || FrameUnsent || Rate > 1.02 || Rate < 0.98) {
      out << stamp(now) << " Statistics - Frames read: " 
	  << FrameCount << " Lost: " << FrameLost << " Unsent: " 
	  << FrameUnsent << " Rate: " << Rate << endl;
   }
   Reset();
   watchdog.clear();
}

void 
Stats::Reset(void) {
   GPSstart   = GPSlast;
   RecIDstart = RecIDlast;
   FrameCount = 0;
   FrameLost  = 0;
   FrameUnsent = 0;
   Stamp = Now();
}

void 
Stats::Start(ulong_t Secs) {
   struct itimerval ticks;

   ticks.it_value.tv_sec  = Secs;
   ticks.it_value.tv_usec  = 0;
   ticks.it_interval.tv_sec  = Secs;
   ticks.it_interval.tv_usec  = 0;
   nSecs = Secs;
   setitimer(ITIMER_REAL, &ticks, 0);
   watchdog.setMode(SigFlag::kBlock);
   watchdog.add(SIGALRM);
   Stamp = Now();
}

void 
Stats::Update(ulong_t FrameID, const Time& GPS, ostream& log) {
   if (!GPSlast) {
      log << stamp(0) << "First frame read, GPS: " << GPS.getS() 
	  << " Frame: " << FrameID << endl;
      GPSlast   = GPS;
      RecIDlast = FrameID;
      Reset();
   } else if (FrameID <= RecIDlast) {
      log << stamp(0) << "Frame received out of order, Frame=" << FrameID
	  << " expected " << RecIDlast+1 << endl;
   } else if (GPS <= GPSlast) {
      log << stamp(0) << "Frame received out of order, GPS=" << GPS.getS() 
	  << " expected " << GPSlast.getS()+1 << endl;
   } else {
      ulong_t lost = FrameID - RecIDlast - 1;
      FrameLost   += lost;
      FrameUnsent += GPS.getS() - GPSlast.getS() - 1 - lost;
   }
   FrameCount++;
   GPSlast   = GPS;
   RecIDlast = FrameID;
}
